import json
import os
import re
import sys
import unicodedata


def config_dir():
    xdg_config_home = os.getenv('XDG_CONFIG_HOME')
    config_home = xdg_config_home if xdg_config_home else os.path.join(os.getenv("HOME"), ".config")

    return os.path.join(config_home, "typobuster")


def get_theme_names():
    theme_dirs = []
    for d in get_data_dirs():
        p = os.path.join(d, "themes")
        if os.path.isdir(p):
            theme_dirs.append(p)

    home = os.getenv("HOME")
    if home:
        p = os.path.join(home, ".themes")
        if os.path.isdir(p):
            theme_dirs.append(p)

    names = []
    exclusions = ["Default", "Emacs"]
    for d in theme_dirs:
        for item in os.listdir(d):
            if os.path.isdir(os.path.join(d, item)) and item not in exclusions:
                content = os.listdir(os.path.join(d, item))
                for name in content:
                    if name.startswith("gtk-"):
                        if item not in names:
                            names.append(item)
                            break
    names.sort()
    return names


def get_data_dirs():
    dirs = [get_data_home()]
    xdg_data_dirs = os.getenv("XDG_DATA_DIRS") if os.getenv("XDG_DATA_DIRS") else "/usr/local/share/:/usr/share/"
    for d in xdg_data_dirs.split(":"):
        dirs.append(d)
    confirmed = []
    for d in dirs:
        if os.path.isdir(d):
            confirmed.append(d)

    return confirmed


def get_data_home():
    d_home = os.getenv('XDG_DATA_HOME') if os.getenv('XDG_DATA_HOME') else os.path.join(
        os.getenv("HOME"), ".local/share")
    return d_home


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def load_json(path):
    try:
        with open(path, 'r') as f:
            return json.load(f)
    except Exception as e:
        eprint("Error loading json: {}".format(e))
        return {}


def save_json(src_dict, path):
    try:
        with open(path, 'w') as f:
            json.dump(src_dict, f, indent=2)
        return "ok"
    except Exception as e:
        return e


def load_settings():
    # check if config dir exists, create if not
    if not os.path.isdir(config_dir()):
        os.makedirs(config_dir())
        print(f"Created {config_dir()}")

    config_path = os.path.join(config_dir(), "config")

    defaults = {
        "auto-indent": True,
        "gspell-enable": False,
        "gspell-lang": "",
        "gtk-font-name": "",
        "gtk-theme-name": "",
        "highlight-current-row": False,
        "highlight-matching-brackets": False,
        "icon-set": "light",
        "icon-size": 24,
        "right-margin-position": 80,
        "right-margin-show": False,
        "sanitize-add-spaces-after-punctuation": True,
        "sanitize-eol": True,
        "sanitize-hyphens": True,
        "sanitize-punctuation-marks": True,
        "sanitize-quotes": True,
        "sanitize-spaces": True,
        "show-bar": True,
        "show-change": False,
        "show-stats": False,
        "syntax": "none",
        "tab-mode": "spaces",
        "tab-width": 4,
        "view-line-numbers": False,
        "whitespaces": False,
        "window-height": 600,
        "window-width": 800,
        "wrap-lines": True
    }
    settings = load_json(config_path)

    # if config file empty or not found
    if not settings:
        result = save_json(defaults, config_path)
        if result == "ok":
            print(f"Saved default settings to {config_path}")
        else:
            eprint(f"Error saving default settings to {config_path}: {result}")

    # add missing keys
    changed = False
    for key in defaults:
        if key not in settings:
            settings[key] = defaults[key]
            changed = True

    if changed:
        result = save_json(settings, config_path)
        if result == "ok":
            print(f"Updated settings in {config_path}")
        else:
            eprint(f"Error updating settings in {config_path}: {result}")

    return settings


def save_settings(settings):
    config_path = os.path.join(config_dir(), "config")
    result = save_json(settings, config_path)
    if result == "ok":
        print(f"Saved settings to {config_path}")
    else:
        eprint(f"Error saving settings to {config_path}: {result}")


def load_syntax():
    syntax_path = os.path.join(config_dir(), "syntax")

    defaults = {
        'c': 'C',
        'cpp': 'C++',
        'css': 'CSS',
        'csv': 'CSV',
        'desktop': '.desktop',
        'diff': 'Diff',
        'go': 'Go',
        'html': 'HTML',
        'java': 'Java',
        'js': 'JavaScript',
        'json': 'JSON',
        'kotlin': 'Kotlin',
        'makefile': 'Makefile',
        'markdown': 'Markdown',
        'meson': 'Meson',
        'php': 'PHP',
        'python3': 'Python',
        'python': 'Python2',
        'r': 'R',
        'sh': 'sh',
        'sql': 'SQL',
        'vala': 'Vala',
        'xml': 'XML',
        'yaml': 'YAML'
    }
    syntax = load_json(syntax_path)

    # if syntax file empty or not found
    if not syntax:
        result = save_json(defaults, syntax_path)
        if result == "ok":
            print(f"Saved default syntax dictionary to {syntax_path}")
        else:
            eprint(f"Error saving default settings to {syntax_path}: {result}")

    # add missing keys
    changed = False
    for key in defaults:
        if key not in syntax:
            syntax[key] = defaults[key]
            changed = True

    if changed:
        result = save_json(syntax, syntax_path)
        if result != "ok":
            eprint(f"Error updating settings in {syntax_path}: {result}")

    return syntax


def load_text_file(path):
    try:
        with open(path, 'r') as file:
            data = file.read()
            return data
    except Exception as e:
        eprint(e)
        return ""


def save_text_file(text, path):
    try:
        with open(path, 'w') as file:
            file.write(text)
            return "ok"
    except Exception as e:
        eprint(e)
        return e


def selected_text(buffer):
    start = buffer.get_start_iter()
    end = buffer.get_end_iter()

    if buffer.get_has_selection():
        try:
            start_iter, end_iter = buffer.get_selection_bounds()
        except:
            return "", 0, 0
    else:
        start_iter = start
        end_iter = end

    return buffer.get_text(start, end, True), start_iter.get_offset(), end_iter.get_offset()


def replace_all(text, old, new):
    return re.sub(re.escape(old), new, text)


def sanitize_hyphens(text, start_idx, end_idx):
    selection = text[start_idx:end_idx]
    selection = selection.replace("–", "-")  # Replace en-dashes with hyphens
    selection = selection.replace(" -", " - ")  # Add spaces around hyphens
    selection = selection.replace("- ", " - ")  # Add spaces around hyphens
    return text[:start_idx] + selection + text[end_idx:]


def sanitize_quotes(text, start_idx, end_idx):
    selection = text[start_idx:end_idx]
    selection = selection.replace(',,', '"')  # Replace double comma with English-style quotes
    selection = re.sub(r"[„”]", '"', selection)  # Replace German-style quotes with English-style quotes
    return text[:start_idx] + selection + text[end_idx:]


def sanitize_punctuation_marks(text, start_idx, end_idx):
    selection = text[start_idx:end_idx]
    selection = re.sub(r'\s+([.,!?;:])', r'\1', selection)  # Remove spaces before punctuation marks
    selection = selection.replace('. ,', '.,')
    return text[:start_idx] + selection + text[end_idx:]


def add_spaces_after_punctuation_marks(text, start_idx, end_idx):
    selection = text[start_idx:end_idx]
    selection = re.sub(r'([.,!?;:])([A-Za-z])', r'\1 \2', selection)  # Add a space after punctuation marks
    return text[:start_idx] + selection + text[end_idx:]


def sanitize_spaces(text, start_idx, end_idx, convert_tabs, tab_width):
    selection = text[start_idx:end_idx]
    selection = re.sub(r" {2,}", " ", selection)  # Replace two or more spaces with a single space
    if convert_tabs:
        selection = re.sub(r'\t+', ' ' * tab_width, selection).strip()  # Replace tabs with a single space
        selection = re.sub(r" {2,}", " ", selection)  # Replace two or more spaces with a single space
    selection = selection.replace(" \n", "\n")  # Remove spaces before end-of-line characters
    selection = selection.replace("\n ", "\n")  # Remove spaces right after end-of-line characters
    return text[:start_idx] + selection + text[end_idx:]


def sanitize_eol(text, start_idx, end_idx):
    selection = text[start_idx:end_idx]
    selection = re.sub(r"\n{2,}", "\n", selection)  # Replace two or more end-of-line characters with a single one
    selection = selection.replace("\n", "\n\n")  # Double all end-of-line characters
    return text[:start_idx] + selection + text[end_idx:]


def get_shell_data_dir():
    data_dir = ""
    home = os.getenv("HOME")
    xdg_data_home = os.getenv("XDG_DATA_HOME")

    if xdg_data_home:
        data_dir = os.path.join(xdg_data_home, "nwg-shell/")
    else:
        if home:
            data_dir = os.path.join(home, ".local/share/nwg-shell/")

    return data_dir


def load_shell_data():
    shell_data_file = os.path.join(get_shell_data_dir(), "data")
    shell_data = load_json(shell_data_file) if os.path.isfile(shell_data_file) else {}

    defaults = {
        "interface-locale": ""
    }

    for key in defaults:
        if key not in shell_data:
            shell_data[key] = defaults[key]

    return shell_data


def to_snake_case(text):
    lines = text.splitlines()
    result = []
    for line in lines:
        line = line.lower().replace(" ", "_")
        result.append(line)
    return "\n".join(result)


def to_kebab_case(text):
    lines = text.splitlines()
    result = []
    for line in lines:
        line = line.lower().replace(" ", "-")
        result.append(line)
    return "\n".join(result)


def to_camel_case(text):
    lines = text.splitlines()
    result = []
    for line in lines:
        line = ''.join(x for x in line.title() if not x.isspace())
        line = line[0].lower() + line[1:]
        result.append(line)
    return "\n".join(result)


def to_upper(text):
    return text.upper()


def to_lower_case(text):
    return text.lower()


def as_in_sentence(text):
    lines = text.splitlines()
    result = []
    for line in lines:
        line = line.lower()
        line = line[0].upper() + line[1:]
        result.append(line)
    return "\n".join(result)


def as_in_title(text):
    lines = text.splitlines()
    modified_lines = []
    for line in lines:
        line = line.lower()
        words = line.split()
        modified_words = []
        for i in range(len(words)):
            word = words[i]
            if i == 0:
                # always capitalize 1st word, even a single letter
                w = word[0].upper() + word[1:]
            else:
                # don't capitalize subsequent words if they are single letters
                if len(word) > 1:
                    w = word[0].upper() + word[1:]
                else:
                    w = word
            modified_words.append(w)
        modified_lines.append(" ".join(modified_words))
    return "\n".join(modified_lines)


def unordered_list(text):
    text = remove_empty_lines(text)
    lines = text.splitlines()
    output = []
    for line in lines:
        line = line.strip().rstrip()
        if not line.startswith('- '):
            line = "- " + line
        output.append(line)

    return "\n".join(output)


def ordered_list(text):
    text = remove_empty_lines(text)
    lines = text.splitlines()
    output = []
    for i in range(len(lines)):
        line = lines[i]
        line = line.strip().rstrip()
        line = f"{i + 1}. {line}"
        output.append(line)

    return "\n".join(output)


def sort_lines(text, order="asc"):
    lines = text.splitlines()
    if order == "asc":
        lines = sorted(lines, key=str.lower)
    else:
        lines = sorted(lines, reverse=True, key=str.lower)
    return "\n".join(lines)


def remove_empty_lines(text):
    return "\n".join(line for line in text.splitlines() if line.strip())


def move_first_word_to_end(text):
    result = []
    lines = text.splitlines()
    for line in lines:
        words = line.split(maxsplit=1)  # Split into first word and the rest
        line = f"{words[1]} {words[0]}" if len(words) > 1 else text
        result.append(line)
    return "\n".join(result)


def remove_non_ascii(text):
    # Normalize text to decompose accents
    normalized = unicodedata.normalize('NFKD', text)

    # Manually replace ł with l and Ł with L
    fixed = normalized.replace('ł', 'l').replace('Ł', 'L')

    # Remove remaining non-ASCII characters
    return ''.join(c for c in fixed if c.encode('ascii', 'ignore'))


def move_last_word_to_beginning(text):
    result = []
    lines = text.splitlines()
    for line in lines:
        words = line.rsplit(maxsplit=1)  # Split into first word and the rest
        line = f"{words[1]} {words[0]}" if len(words) > 1 else text
        result.append(line)
    return "\n".join(result)


def merge_lines(text):
    lines = text.splitlines()
    return " ".join(lines)



"""
Lightweight editor with text transformations and auto-correction.
Project: https://github.com/nwg-piotr/typobuster
Author's email: nwg.piotr@gmail.com
Copyright (c) 2025 Piotr Miller & Contributors
License: GPL3
"""

import argparse
import cairo
import os.path
import subprocess

import gi

gi.require_version("Gtk", "3.0")
gi.require_version("GtkSource", "4")
from gi.repository import Gtk, Gdk, GLib, GtkSource

from typobuster.ui_components import MenuBar, ButtonBar, SanitizationDialog, AboutWindow, SearchBar, PreferencesDialog
from typobuster.tools import *

from typobuster.__about__ import __version__

dir_name = os.path.dirname(__file__)
file_path = ""
voc = {}


def load_vocabulary():
    global voc
    # basic vocabulary (for en_US)
    voc = load_json(os.path.join(dir_name, "langs", "en_US.json"))
    if not voc:
        eprint("Failed loading vocabulary, terminating")
        sys.exit(1)

    shell_data = load_shell_data()

    lang = os.getenv("LANG")
    if lang is None:
        lang = "en_US"
    else:
        lang = lang.split(".")[0] if not shell_data["interface-locale"] else shell_data["interface-locale"]

    # translate if translation available
    if lang != "en_US":
        loc_file = os.path.join(dir_name, "langs", "{}.json".format(lang))
        if os.path.isfile(loc_file):
            # localized vocabulary
            loc = load_json(loc_file)
            if not loc:
                eprint("Failed loading translation into '{}'".format(lang))
            else:
                for key in loc:
                    voc[key] = loc[key]


def on_destroy_event(widget):
    print("Terminating gracefully")
    Gtk.main_quit()


class Typobuster(Gtk.Window):
    def __init__(self):
        super().__init__()
        self.button_bar = None
        self.drag_in_progress = False
        self.settings = load_settings()
        self.syntax_dict = load_syntax()

        if self.settings["window-width"] and self.settings["window-height"]:
            self.set_default_size(self.settings["window-width"], self.settings["window-height"])
        else:
            self.set_default_size(800, 600)

        self.voc = voc
        self.set_title(f"{voc['untitled']} - Typobuster")

        self.last_dir_path = ""
        self.search_bar = None

        self.gtk_settings = Gtk.Settings.get_default()

        self.initial_text = ""
        self.file_stat = None

        self.gspell_available = False
        self.gspell_text_view = None

        self.connect("delete-event", self.on_close)
        self.connect("key-release-event", self.handle_keyboard_release)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=0)
        self.add(vbox)

        # Create a GtkSourceView and configure it
        self.source_view = GtkSource.View()

        self.source_view.get_style_context().add_class("sourceview")
        if self.settings["view-line-numbers"]:
            self.source_view.set_show_line_numbers(True)  # Enable line numbers

        self.source_view.set_right_margin_position(self.settings["right-margin-position"])
        self.source_view.set_show_right_margin(self.settings["right-margin-show"])

        space_drawer = self.source_view.get_space_drawer()
        space_drawer.set_enable_matrix(self.settings["whitespaces"])  # Whether to enable whitespaces view

        self.source_view.set_highlight_current_line(
            self.settings["highlight-current-row"])  # Highlight the current line

        if self.settings["wrap-lines"]:
            self.source_view.set_wrap_mode(Gtk.WrapMode.WORD)
        else:
            self.source_view.set_wrap_mode(Gtk.WrapMode.NONE)

        self.set_tab_width()

        self.set_tab_mode()

        self.set_auto_indent()

        # Enable drag-and-drop for URI (file paths)
        self.source_view.drag_dest_set(Gtk.DestDefaults.ALL, [], Gdk.DragAction.COPY)
        self.source_view.drag_dest_add_uri_targets()
        self.source_view.connect("drag-data-received", self.on_drag_data_received)

        # Set a language for syntax highlighting
        self.lang_manager = GtkSource.LanguageManager()
        self.buffer = GtkSource.Buffer()

        self.buffer.set_highlight_matching_brackets(
            self.settings["highlight-matching-brackets"])  # Highlight matching brackets

        self.source_view.set_buffer(self.buffer)
        self.buffer.connect("changed", self.on_text_changed)
        self.buffer.connect("mark-set", self.on_cursor_moved)

        self.clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)

        self.menu_bar = MenuBar(self)
        vbox.pack_start(self.menu_bar, False, False, 0)

        self.button_bar_wrapper = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self.button_bar_wrapper.set_property("name", "bar-wrapper")
        vbox.pack_start(self.button_bar_wrapper, False, False, 0)

        self.create_button_bar()

        # Create a scrollable window and add the source view
        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_hexpand(True)
        scrolled_window.set_vexpand(True)
        scrolled_window.add(self.source_view)

        self.connect("enter-notify-event", self.check_file_change)

        # Add the scrollable window to the main window
        vbox.add(scrolled_window)

        self.search_bar = SearchBar(self)
        vbox.pack_end(self.search_bar, False, False, 0)

        if "syntax" in self.settings:
            language = self.lang_manager.get_language(self.settings["syntax"])
            self.buffer.set_language(language)
            if self.settings["syntax"] == "none":
                s_lbl = self.voc["plain-text"]
            else:
                s_lbl = self.syntax_dict[self.settings["syntax"]]
            self.search_bar.syntax_lbl.set_text(s_lbl)

        # Add initial (empty) text to the buffer
        self.buffer.begin_not_undoable_action()
        self.buffer.set_text("")
        self.reset_initial_text()
        self.update_stats()
        self.buffer.end_not_undoable_action()

        try:
            gi.require_version("Gspell", "1")
            from gi.repository import Gspell
            self.gspell_available = True
        except:
            self.gspell_available = False
            print(f"Gspell is NOT installed, spell check unavailable")

        if self.gspell_available:
            # Initialize gspell
            self.gspell_text_view = Gspell.TextView.get_from_gtk_text_view(self.source_view)

            language = None
            # set language from the value in settings
            if self.settings["gspell-lang"]:
                language = Gspell.Language.lookup(self.settings["gspell-lang"])
            # if language unset or erroneous, use the system default
            if not language:
                language = Gspell.Language.get_default()

            # the above might have not worked if hunspell missing
            if language:
                self.checker = Gspell.Checker.new(language)
                print(f"Spell check language: {language.get_code()}")
                self.checker.set_language(language)

                buffer = Gspell.TextBuffer.get_from_gtk_text_buffer(self.buffer)
                buffer.set_spell_checker(self.checker)
                self.gspell_text_view.set_enable_language_menu(True)
                self.gspell_available = True
                self.gspell_text_view.set_inline_spell_checking(
                    self.settings["gspell-enable"] and self.gspell_text_view)
                print("Loaded Gspell module")
            else:
                self.gspell_available = False
                print("Couldn't initialize spell check. Are hunspell, hunspell-en_us, hspell, nuspell, aspell and "
                      "libvoikko installed?")

        self.set_view_style()
        self.set_gtk_theme()

        # Connect the delete event to quit the application
        self.connect("destroy", on_destroy_event)

        self.source_view.grab_focus()

    def create_button_bar(self):
        if self.settings["show-bar"]:
            # we need to destroy and build from scratch in case icon size changed
            if len(self.button_bar_wrapper.get_children()) > 0:
                self.button_bar_wrapper.get_children()[0].destroy()
            self.button_bar = ButtonBar(self, dir_name)
            self.button_bar_wrapper.pack_start(self.button_bar, False, False, 0)
            self.button_bar_wrapper.show_all()
        else:
            self.button_bar_wrapper.hide()

    def text_changed(self):
        return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter(), True) != self.initial_text

    def reset_initial_text(self):
        self.initial_text = self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter(), True)

    def on_text_changed(self, buffer):
        self.update_cursor_position()
        self.update_stats()

    def update_cursor_position(self):
        """ Get and print the cursor position. """
        insert_mark = self.buffer.get_insert()
        iter_ = self.buffer.get_iter_at_mark(insert_mark)
        line = iter_.get_line()
        column = iter_.get_line_offset()
        self.search_bar.pos_lbl.set_text(f'{self.voc["row"]}: {line + 1} {self.voc["column"]}: {column}')

    def on_cursor_moved(self, buffer, iter_, mark):
        """ Fires when the cursor moves. """
        if mark == buffer.get_insert():  # Ignore selection mark
            self.update_cursor_position()
        else:
            if self.settings["show-stats"]:
                self.update_stats()

    def update_stats(self):
        if self.settings["show-stats"]:
            txt, s, e = selected_text(self.buffer)
            selection = txt[s:e]
            self.search_bar.stat_lbl.set_text(
                f'{self.voc["characters"]}: {len(selection)} {self.voc["words"]}: {len(selection.split())}')

        self.mark_changes_in_ui()

    def mark_changes_in_ui(self):
        if self.text_changed():
            if not self.get_title().startswith("*"):
                self.set_window_title(f"*{self.get_title()}")
                self.search_bar.change_lbl.set_text("*")
        else:
            if self.get_title().startswith("*"):
                self.set_window_title(self.get_title()[1:])
                self.search_bar.change_lbl.set_text("")

    def switch_stats_visibility(self):
        if self.search_bar:
            self.search_bar.stat_lbl.set_visible(self.settings["show-stats"])

    def switch_change_visibility(self):
        if self.search_bar:
            self.search_bar.change_lbl.set_visible(self.settings["show-change"])

    def handle_keyboard_release(self, widget, event):
        if event.keyval == Gdk.KEY_Escape:
            if self.search_bar.search_entry.is_focus() and self.search_bar.search_entry.get_text():
                self.search_bar.search_entry.set_text("")
            else:
                self.source_view.grab_focus()
        elif event.keyval == Gdk.KEY_n and event.state & Gdk.ModifierType.CONTROL_MASK:
            self.new_file()
        elif event.keyval == Gdk.KEY_o and event.state & Gdk.ModifierType.CONTROL_MASK:
            self.open_file()
        elif event.keyval == Gdk.KEY_s and event.state & Gdk.ModifierType.CONTROL_MASK:
            self.save_file()
        elif event.keyval == Gdk.KEY_S and (event.state & Gdk.ModifierType.CONTROL_MASK) and (
                event.state & Gdk.ModifierType.SHIFT_MASK):
            self.save_file_as()
        elif event.keyval == Gdk.KEY_p and event.state & Gdk.ModifierType.CONTROL_MASK:
            self.on_print_btn()
        elif event.keyval == Gdk.KEY_q and event.state & Gdk.ModifierType.CONTROL_MASK:
            self.quit()
        elif event.keyval == Gdk.KEY_z and event.state & Gdk.ModifierType.CONTROL_MASK:
            self.undo()
        elif event.keyval == Gdk.KEY_y and event.state & Gdk.ModifierType.CONTROL_MASK:
            self.redo()
        elif event.keyval == Gdk.KEY_f and event.state & Gdk.ModifierType.CONTROL_MASK:
            self.search_selection()

    def search_selection(self):
        txt, s, e = selected_text(self.buffer)
        selection = txt[s:e]
        if not (s == 0 and e == len(txt)):
            self.search_bar.search_entry.set_text(selection)
        else:
            self.search_bar.search_entry.set_text("")

        self.search_bar.search_entry.grab_focus()

    def on_close(self, widget, event):
        # remember last window dimensions
        size = self.get_size()
        self.settings["window-width"] = size.width
        self.settings["window-height"] = size.height

        # remember last used spell check language
        if self.gspell_available and self.settings["gspell-lang"] != self.checker.get_language().get_code():
            print(f"Storing changed spell check language: {self.checker.get_language().get_code()}")
            self.settings["gspell-lang"] = self.checker.get_language().get_code()

        save_settings(self.settings)

        if self.text_changed():
            dialog = Gtk.MessageDialog(
                transient_for=self,
                flags=0,
                message_type=Gtk.MessageType.WARNING,

                text=self.voc["unsaved-changes"]
            )
            dialog.add_buttons(self.voc["save"], Gtk.ResponseType.YES, self.voc["dont-save"], Gtk.ResponseType.NO,
                               self.voc["cancel"], Gtk.ResponseType.CANCEL)
            dialog.format_secondary_text(self.voc["unsaved-changes-question"])
            response = dialog.run()
            dialog.destroy()

            if response == Gtk.ResponseType.YES:
                self.save_on_exit()
                return False  # Allow closing
            elif response == Gtk.ResponseType.NO:
                return False  # Allow closing
            else:
                return True  # Prevent closing
        return False  # Allow closing

    def save_on_exit(self):
        if self.get_title() == f"{voc['untitled']} - Typobuster":
            self.save_file_as(None)
        else:
            self.save_file(None)

    def set_syntax(self, widget, name):
        language = self.lang_manager.get_language(name)
        self.buffer.set_language(language)
        self.settings["syntax"] = name
        save_settings(self.settings)
        if self.settings["syntax"] == "none":
            s_lbl = self.voc["plain-text"]
        else:
            s_lbl = self.syntax_dict[self.settings["syntax"]]
        self.search_bar.syntax_lbl.set_text(s_lbl)

    def undo(self, *args):
        if self.buffer.can_undo():
            self.buffer.undo()

    def redo(self, *args):
        if self.buffer.can_redo():
            self.buffer.redo()

    def cut_text(self, *args):
        self.buffer.begin_user_action()
        self.buffer.cut_clipboard(self.clipboard, True)  # True -> delete after copying
        self.buffer.end_user_action()

    def copy_text(self, widget):
        self.buffer.copy_clipboard(self.clipboard)

    def paste_text(self, widget):
        self.buffer.begin_user_action()
        self.buffer.paste_clipboard(self.clipboard, None, True)
        self.buffer.end_user_action()

    def delete_text(self, widget):
        self.buffer.begin_user_action()
        self.buffer.delete_selection(True, True)
        self.buffer.end_user_action()

    def toggle_line_numbers(self, widget):
        self.settings["view-line-numbers"] = widget.get_active()
        self.source_view.set_show_line_numbers(self.settings["view-line-numbers"])
        save_settings(self.settings)

    def toggle_whitespaces(self, widget):
        self.settings["whitespaces"] = widget.get_active()
        space_drawer = self.source_view.get_space_drawer()
        space_drawer.set_enable_matrix(self.settings["whitespaces"])
        save_settings(self.settings)

    def toggle_highlight_current_row(self, widget):
        self.settings["highlight-current-row"] = widget.get_active()
        self.source_view.set_highlight_current_line(self.settings["highlight-current-row"])
        save_settings(self.settings)

    def toggle_highlight_matching_brackets(self, widget):
        self.settings["highlight-matching-brackets"] = widget.get_active()
        self.source_view.get_buffer().set_highlight_matching_brackets(self.settings["highlight-matching-brackets"])
        save_settings(self.settings)

    def toggle_line_wrap(self, widget):
        self.settings["wrap-lines"] = widget.get_active()
        if self.settings["wrap-lines"]:
            self.source_view.set_wrap_mode(Gtk.WrapMode.WORD)
        else:
            self.source_view.set_wrap_mode(Gtk.WrapMode.NONE)
        save_settings(self.settings)

    def show_about(self, widget):
        about = AboutWindow(self)
        about.run()

    def show_preferences(self, widget):
        PreferencesDialog(self)

    def on_font_selected(self, font_btn):
        self.settings["gtk-font-name"] = font_btn.get_font()
        self.set_view_style()
        save_settings(self.settings)

    def on_tab_with_selected(self, sb):
        self.settings["tab-width"] = int(sb.get_value())
        self.set_tab_width()
        save_settings(self.settings)

    def on_margin_position_selected(self, sb):
        self.settings["right-margin-position"] = int(sb.get_value())
        self.source_view.set_right_margin_position(self.settings["right-margin-position"])
        save_settings(self.settings)

    def on_icon_size_selected(self, sb):
        self.settings["icon-size"] = int(sb.get_value())
        save_settings(self.settings)
        self.create_button_bar()

    def on_right_margin_toggled(self, cb):
        self.settings["right-margin-show"] = cb.get_active()
        self.source_view.set_show_right_margin(self.settings["right-margin-show"])
        save_settings(self.settings)

    def on_bar_show_toggled(self, cb):
        self.settings["show-bar"] = cb.get_active()
        save_settings(self.settings)
        self.create_button_bar()

    def on_tab_mode_changed(self, combo):
        self.settings["tab-mode"] = combo.get_active_id()
        self.set_tab_mode()
        save_settings(self.settings)

    def on_icon_set_changed(self, combo):
        self.settings["icon-set"] = combo.get_active_id()
        save_settings(self.settings)
        self.create_button_bar()

    def on_auto_indent_changed(self, check_button):
        self.settings["auto-indent"] = check_button.get_active()
        self.set_auto_indent()
        save_settings(self.settings)

    def on_spell_check_switched(self, check_button):
        self.settings["gspell-enable"] = check_button.get_active()
        if self.gspell_available and self.gspell_text_view:
            self.gspell_text_view.set_inline_spell_checking(self.settings["gspell-enable"])
        save_settings(self.settings)

    def on_stats_cb_toggled(self, check_button):
        self.settings["show-stats"] = check_button.get_active()
        self.switch_stats_visibility()
        save_settings(self.settings)
        self.update_stats()

    def on_change_cb_toggled(self, check_button):
        self.settings["show-change"] = check_button.get_active()
        self.switch_change_visibility()
        save_settings(self.settings)
        self.update_stats()

    def set_view_style(self):
        if self.settings["gtk-font-name"]:
            # Create a CssProvider and parse "gtk-font-name" into CSS
            css_provider = Gtk.CssProvider()
            font = self.settings["gtk-font-name"].replace(",", "")
            parts = font.split()
            family = " ".join(parts[0:-1])
            size = parts[-1]
            css = f'.sourceview {{ font-family: "{family}"; font-size: {size}pt; }}'.encode()
            css_provider.load_from_data(css)

            # Get the style context of self.source_view and add the CSS provider
            style_context = self.source_view.get_style_context()
            style_context.add_provider(css_provider, Gtk.STYLE_PROVIDER_PRIORITY_USER)

    def set_tab_width(self):
        self.source_view.set_tab_width(self.settings["tab-width"])

    def set_tab_mode(self):
        self.source_view.set_insert_spaces_instead_of_tabs(self.settings["tab-mode"] == "spaces")

    def set_auto_indent(self):
        self.source_view.set_auto_indent(self.settings["auto-indent"])

    def on_theme_changed(self, combo):
        self.settings["gtk-theme-name"] = combo.get_active_id()
        save_settings(self.settings)
        self.set_gtk_theme()

    def set_gtk_theme(self):
        if self.settings["gtk-theme-name"]:
            self.gtk_settings.set_property("gtk-theme-name", self.settings["gtk-theme-name"])
        else:
            theme = subprocess.check_output("gsettings get org.gnome.desktop.interface gtk-theme", shell=True).decode(
                "utf-8")
            self.gtk_settings.set_property("gtk-theme-name", theme[1:-2])

    def on_drag_data_received(self, widget, drag_context, x, y, data, info, time):
        # workaround for drag-data-received being fired twice
        if self.drag_in_progress:
            return
        self.drag_in_progress = True

        """Handle file drop event."""
        uris = data.get_uris()
        if uris:
            f_p = uris[0].replace("file://", "").strip()
            self.load_file(None, f_p)
            global file_path
            file_path = f_p
        self.drag_in_progress = False

    def transform_text(self, widget, transformation):
        start = self.buffer.get_start_iter()
        end = self.buffer.get_end_iter()
        if self.buffer.get_has_selection():
            start, end = self.buffer.get_selection_bounds()
        if start and end:
            text = self.buffer.get_text(start, end, True)

            if transformation == "sentence":
                transformed_text = as_in_sentence(text)
            elif transformation == "title":
                transformed_text = as_in_title(text)
            elif transformation == "lowercase":
                transformed_text = to_lower_case(text)
            elif transformation == "uppercase":
                transformed_text = to_upper(text)
            elif transformation == "camelcase":
                transformed_text = to_camel_case(text)
            elif transformation == "snakecase":
                transformed_text = to_snake_case(text)
            elif transformation == "kebabcase":
                transformed_text = to_kebab_case(text)
            elif transformation == "unordered":
                transformed_text = unordered_list(text)
            elif transformation == "ordered":
                transformed_text = ordered_list(text)
            elif transformation == "sort-asc":
                transformed_text = sort_lines(text)
            elif transformation == "sort-desc":
                transformed_text = sort_lines(text, order="desc")
            elif transformation == "remove-empty-rows":
                transformed_text = remove_empty_lines(text)
            elif transformation == "remove-non-ascii":
                transformed_text = remove_non_ascii(text)
            elif transformation == "first-to-end":
                transformed_text = move_first_word_to_end(text)
            elif transformation == "last-to-beginning":
                transformed_text = move_last_word_to_beginning(text)
            elif transformation == "merge-rows":
                transformed_text = merge_lines(text)
            else:
                transformed_text = text

            self.buffer.begin_user_action()
            self.buffer.delete(start, end)
            self.buffer.insert(start, transformed_text)
            self.buffer.end_user_action()

    def select_range(self, start, end):
        start_iter = self.buffer.get_iter_at_offset(start)
        end_iter = self.buffer.get_iter_at_offset(end)
        self.buffer.select_range(start_iter, end_iter)

    def sanitize_text(self, widget):
        # Apply some basic predefined sanitization
        d = SanitizationDialog(self, self.buffer)

    def replace(self, old, new):
        start = self.buffer.get_start_iter()
        end = self.buffer.get_end_iter()

        text = self.buffer.get_text(start, end, True)
        text = replace_all(text, old, new)
        self.update_text(text)

    def set_window_title(self, path):
        filename = os.path.basename(path)
        self.set_title(filename)

    def new_file(self, *args):
        title = file_path.split("/")[-1] if file_path else self.voc["untitled"]
        if self.text_changed():
            if self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter(), True):
                dialog = Gtk.MessageDialog(
                    transient_for=self,
                    flags=0,
                    message_type=Gtk.MessageType.QUESTION,
                    buttons=Gtk.ButtonsType.YES_NO,
                    text=f"{self.voc['want-save-changes']} {title}?",
                )
                dialog.format_secondary_text(self.voc["chages-will-be-lost"])
                response = dialog.run()
                dialog.destroy()
                if response == Gtk.ResponseType.YES:
                    self.save_file(self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter(), True))

        self.buffer.begin_not_undoable_action()
        self.update_text("")
        self.reset_initial_text()
        self.buffer.end_not_undoable_action()
        self.update_stats()
        self.set_window_title(f"{voc['untitled']} - Typobuster")

    def load_file(self, widget, path):
        if self.text_changed():
            resp = self.on_close(None, None)
            if resp:
                return  # Prevent from closing

        global file_path
        file_path = os.path.abspath(path)
        if os.path.isfile(file_path):
            print(f"Opening {file_path}")
        else:
            eprint(f"'{file_path}' does not exist")

            dialog = Gtk.MessageDialog(
                parent=self,
                transient_for=self,
                flags=0,
                message_type=Gtk.MessageType.ERROR,
                buttons=Gtk.ButtonsType.OK,
                text=voc["file-not-found"],
            )
            dialog.run()
            dialog.destroy()

            self.update_recent(file_path, remove=True)
            return

        text = ""
        if os.path.isfile(path):
            self.last_dir_path = os.path.dirname(path)  # remember last opened dir for file chooser
            text = load_text_file(path)
            self.file_stat = os.stat(file_path)
            self.update_recent(path)
            self.menu_bar.recent_menu_item.set_sensitive(True)
        self.buffer.begin_not_undoable_action()
        self.update_text(text)
        self.reset_initial_text()
        self.buffer.end_not_undoable_action()
        self.update_stats()
        self.set_window_title(path)

        self.search_bar.clear()
        self.source_view.grab_focus()

    def open_file(self, *args):
        if self.text_changed():
            if self.on_close(None, None):
                return  # Prevent from closing

        dialog = Gtk.FileChooserDialog(
            title="Open File",
            parent=self,
            action=Gtk.FileChooserAction.OPEN,
        )
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OPEN, Gtk.ResponseType.OK)
        if self.last_dir_path:
            dialog.set_current_folder(self.last_dir_path)
        else:
            dialog.set_current_folder(os.getenv("HOME"))

        filter_text = Gtk.FileFilter()
        filter_text.set_name("Text files")
        filter_text.add_mime_type("text/plain")
        dialog.add_filter(filter_text)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            filename = dialog.get_filename()
            self.load_file(None, filename)
        dialog.destroy()

    def save_file(self, *args):
        if file_path:
            text = self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter(), True)
            result = save_text_file(text, file_path)
            if result == "ok":
                print(f"Saved text to {file_path}")
                self.update_recent(file_path)
                self.file_stat = os.stat(file_path)
                self.reset_initial_text()
                self.update_stats()
            else:
                eprint(f"Error saving text to {file_path}: {result}")
        else:
            self.save_file_as(None)

    def save_file_as(self, *args):
        global file_path
        dialog = Gtk.FileChooserDialog(
            title="Save File",
            parent=self,
            action=Gtk.FileChooserAction.SAVE,
        )
        dialog.add_buttons(Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_SAVE, Gtk.ResponseType.OK)
        if file_path:
            dialog.set_filename(file_path)
        else:
            dialog.set_current_folder(os.getenv("HOME"))

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            filename = dialog.get_filename()
            text = self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter(), True)
            result = save_text_file(text, filename)
            if result == "ok":
                print(f"Saved text to {filename}")
                file_path = filename
                self.set_window_title(filename)
                self.file_stat = os.stat(file_path)
                self.update_recent(file_path)
                self.reset_initial_text()
                self.update_stats()
            else:
                eprint(f"Error saving text to {filename}: {result}")
        dialog.destroy()

    def check_file_change(self, widget, event):
        if self.file_stat:
            current_stat = os.stat(file_path)
            if current_stat.st_size != self.file_stat.st_size or current_stat.st_mtime != self.file_stat.st_mtime:
                dialog = Gtk.MessageDialog(
                    parent=self,
                    message_type=Gtk.MessageType.QUESTION,
                    buttons=Gtk.ButtonsType.YES_NO,
                    text=self.voc["file-changed-externally"],
                )
                dialog.format_secondary_text(self.voc["reload-file"])

                response = dialog.run()
                dialog.destroy()

                if response == Gtk.ResponseType.YES:
                    self.load_file(None, file_path)
                    self.file_stat = os.stat(file_path)

                self.file_stat = os.stat(file_path)

    def update_text(self, text):
        start, end = self.buffer.get_bounds()
        self.buffer.begin_user_action()  # Mark as a single undoable action
        self.buffer.delete(start, end)
        self.buffer.insert_at_cursor(text)
        self.buffer.end_user_action()  # End the undoable action

    def update_recent(self, path, remove=False):
        recent_file = os.path.join(config_dir(), "recent")
        recent_paths = load_text_file(recent_file).splitlines() if os.path.isfile(recent_file) else []
        if path in recent_paths:
            recent_paths.remove(path)
        if len(recent_paths) >= 10:
            last = recent_paths.pop()
            print(f"Removing from recent: '{last}'")
        if not remove:
            recent_paths.insert(0, path)
        save_text_file("\n".join(recent_paths), recent_file)

    def on_print_btn(self, *args):
        print_settings = Gtk.PrintSettings()
        page_setup = Gtk.PageSetup()

        print_operation = Gtk.PrintOperation()
        print_operation.set_print_settings(print_settings)
        print_operation.set_default_page_setup(page_setup)
        print_operation.connect("begin-print", self.begin_print)
        print_operation.connect("draw-page", self.draw_page)

        result = print_operation.run(Gtk.PrintOperationAction.PRINT_DIALOG, self)
        if result == Gtk.PrintOperationResult.APPLY:
            print("Printing in progress...")
        elif result == Gtk.PrintOperationResult.CANCEL:
            print("Print canceled")
        elif result == Gtk.PrintOperationResult.ERROR:
            print("Print error occurred")
        else:
            print(f"Unknown result: {result}")

    def begin_print(self, print_operation, context):
        print("Begin print")
        self.pages = self.paginate_text(context)
        print_operation.set_n_pages(len(self.pages))

    def draw_page(self, print_operation, context, page_nr):
        print("Drawing page")
        cr = context.get_cairo_context()
        cr.set_source_rgb(0, 0, 0)
        cr.select_font_face("Sans", cairo.FONT_SLANT_NORMAL, cairo.FONT_WEIGHT_NORMAL)
        cr.set_font_size(10)

        lines = self.pages[page_nr]
        line_height = 14
        y = 10

        for line in lines:
            cr.move_to(10, y)
            cr.show_text(line)
            y += line_height

    def wrap_text(self, cr, text, max_width):
        lines = []
        paragraphs = text.split('\n')

        for paragraph in paragraphs:
            if paragraph.strip() == "":
                lines.append("")  # Preserve empty line
                continue

            words = paragraph.split()
            current_line = []

            for word in words:
                current_line.append(word)
                line_width = cr.text_extents(" ".join(current_line)).width
                if line_width > max_width:
                    current_line.pop()
                    lines.append(" ".join(current_line))
                    current_line = [word]

            if current_line:
                lines.append(" ".join(current_line))

        return lines

    def paginate_text(self, context):
        cr = context.get_cairo_context()
        text = self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter(), True)
        lines = self.wrap_text(cr, text, 450)  # Adjust according to the actual page width in points
        line_height = 14
        page_height = 800  # Adjust according to the actual page height
        max_lines_per_page = int(page_height / line_height)

        pages = []
        current_page = []

        for line in lines:
            if len(current_page) >= max_lines_per_page:
                pages.append(current_page)
                current_page = []

            current_page.append(line)

        if current_page:
            pages.append(current_page)

        return pages

    def quit(self, *args):
        self.close()


def main():
    parser = argparse.ArgumentParser(description="Simple text editor")
    parser.add_argument("file_path", type=str, nargs="?", help="Path of the file to open")
    parser.add_argument("-v",
                        "--version",
                        action="version",
                        version="%(prog)s version {}".format(__version__),
                        help="display version information")

    args = parser.parse_args()

    if args.file_path:
        global file_path
        file_path = args.file_path

    GLib.set_prgname('typobuster')
    load_vocabulary()

    window = Typobuster()

    screen = Gdk.Screen.get_default()
    provider = Gtk.CssProvider()
    style_context = Gtk.StyleContext()
    style_context.add_provider_for_screen(screen, provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
    css = b"""
        #searchentry-error {
            color: red;
        }
        #bar-button {
            min-width: 16px;
            min-height: 16px;
            background: none;
            border: none;
            padding: 0px;
        }
        #bar-button:hover {
            background-color: rgba(0, 255, 255, 0.5);
        }
        #bar-wrapper {
            padding: 3px;
        }
        """
    provider.load_from_data(css)

    if args.file_path:
        window.load_file(None, args.file_path)

    window.show_all()
    window.switch_stats_visibility()
    window.switch_change_visibility()
    Gtk.main()


# Run the application
if __name__ == "__main__":
    sys.exit(main())



import os.path

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Gdk, GdkPixbuf

from typobuster.tools import *
from typobuster.__about__ import __version__


class MenuBar(Gtk.MenuBar):
    def __init__(self, parent_window):
        super().__init__()
        self.settings = parent_window.settings
        self.set_take_focus(False)

        self.parent_window = parent_window

        # File menu
        file_menu = Gtk.Menu()
        file_menu.connect("show", self.update_menu_items_sensitivity)
        file_menu_item = Gtk.MenuItem(label=parent_window.voc["file"])
        file_menu_item.set_submenu(file_menu)

        # File/New
        self.new_menu_item = CustomMenuItem(parent_window.voc["new"], "Ctrl+N")
        file_menu.append(self.new_menu_item)
        self.new_menu_item.connect("activate", parent_window.new_file)

        # File/Open
        open_menu_item = CustomMenuItem(parent_window.voc["open"], "Ctrl+O")
        file_menu.append(open_menu_item)
        open_menu_item.connect("activate", parent_window.open_file)

        # File/Open recent
        self.recent_menu_item = Gtk.MenuItem(label=parent_window.voc["recent-files"])
        file_menu.append(self.recent_menu_item)
        self.recent_menu_item.set_sensitive(os.path.isfile(os.path.join(config_dir(), "recent")))
        file_menu.connect("show", add_recent_menu, self.recent_menu_item, self.parent_window)

        # File/Save
        save_menu_item = CustomMenuItem(parent_window.voc["save"], "Ctrl+S")
        file_menu.append(save_menu_item)
        save_menu_item.connect("activate", parent_window.save_file)

        # File/Save As
        save_as_menu_item = CustomMenuItem(parent_window.voc["save-as"], "Ctrl+Shift+S")
        file_menu.append(save_as_menu_item)
        save_as_menu_item.connect("activate", parent_window.save_file_as)

        sep = Gtk.SeparatorMenuItem()
        file_menu.append(sep)

        # File/Print
        print_menu_item = CustomMenuItem(parent_window.voc["print"], "Ctrl+P")
        file_menu.append(print_menu_item)
        print_menu_item.connect("activate", parent_window.on_print_btn)

        sep = Gtk.SeparatorMenuItem()
        file_menu.append(sep)

        # Quit menu item
        quit_menu_item = CustomMenuItem(parent_window.voc["quit"], "Ctrl+Q")
        file_menu.append(quit_menu_item)
        quit_menu_item.connect("activate", parent_window.quit)

        # Edit menu
        edit_menu = Gtk.Menu()
        edit_menu.connect("show", self.update_menu_items_sensitivity)
        edit_menu_item = Gtk.MenuItem(label=parent_window.voc["edit"])
        edit_menu_item.set_submenu(edit_menu)

        # Edit/Undo
        self.undo_menu_item = CustomMenuItem(parent_window.voc["undo"], "Ctrl+Z")
        edit_menu.append(self.undo_menu_item)
        self.undo_menu_item.connect("activate", parent_window.undo)

        # Edit/Redo
        self.redo_menu_item = CustomMenuItem(parent_window.voc["redo"], "Ctrl+Y")
        edit_menu.append(self.redo_menu_item)
        self.redo_menu_item.connect("activate", parent_window.redo)

        # Edit/Copy
        copy_menu_item = CustomMenuItem(parent_window.voc["copy"], "Ctrl+C")
        edit_menu.append(copy_menu_item)
        copy_menu_item.connect("activate", parent_window.copy_text)

        # Edit/Cut
        cut_menu_item = CustomMenuItem(parent_window.voc["cut"], "Ctrl+X")
        edit_menu.append(cut_menu_item)
        cut_menu_item.connect("activate", parent_window.cut_text)

        # Edit/Paste
        paste_menu_item = CustomMenuItem(parent_window.voc["paste"], "Ctrl+V")
        edit_menu.append(paste_menu_item)
        paste_menu_item.connect("activate", parent_window.paste_text)

        # Edit/Delete
        delete_menu_item = CustomMenuItem(parent_window.voc["delete"], "Delete")
        edit_menu.append(delete_menu_item)
        delete_menu_item.connect("activate", parent_window.delete_text)

        # Transform menu item
        transform_menu_item = Gtk.MenuItem(label=parent_window.voc["transform"])
        edit_menu.append(transform_menu_item)

        # Transform/As in sentence
        transform_as_in_sentence_menu_item = Gtk.MenuItem(label=parent_window.voc["as-in-sentence"])
        transform_menu_item.set_submenu(Gtk.Menu())
        transform_menu_item.get_submenu().append(transform_as_in_sentence_menu_item)
        transform_as_in_sentence_menu_item.connect("activate", parent_window.transform_text, "sentence")

        # Transform/As In Title
        transform_as_in_title_menu_item = Gtk.MenuItem(label=parent_window.voc["as-in-title"])
        transform_menu_item.get_submenu().append(transform_as_in_title_menu_item)
        transform_as_in_title_menu_item.connect("activate", parent_window.transform_text, "title")

        # Transform/To Uppercase
        transform_uppercase_menu_item = Gtk.MenuItem(label=parent_window.voc["uppercase"])
        transform_menu_item.get_submenu().append(transform_uppercase_menu_item)
        transform_uppercase_menu_item.connect("activate", parent_window.transform_text, "uppercase")

        # Transform/to lowercase
        transform_lowercase_menu_item = Gtk.MenuItem(label=parent_window.voc["lowercase"])
        transform_menu_item.get_submenu().append(transform_lowercase_menu_item)
        transform_lowercase_menu_item.connect("activate", parent_window.transform_text, "lowercase")

        # Transform/toCamelCase
        transform_camelcase_menu_item = Gtk.MenuItem(label=parent_window.voc["camel-case"])
        transform_menu_item.get_submenu().append(transform_camelcase_menu_item)
        transform_camelcase_menu_item.connect("activate", parent_window.transform_text, "camelcase")

        # Transform/to_snake_case
        transform_snakecase_menu_item = Gtk.MenuItem(label=parent_window.voc["snake-case"])
        transform_menu_item.get_submenu().append(transform_snakecase_menu_item)
        transform_snakecase_menu_item.connect("activate", parent_window.transform_text, "snakecase")

        # Transform/to-kebab-case
        transform_kebabcase_menu_item = Gtk.MenuItem(label=parent_window.voc["kebab-case"])
        transform_menu_item.get_submenu().append(transform_kebabcase_menu_item)
        transform_kebabcase_menu_item.connect("activate", parent_window.transform_text, "kebabcase")

        # Transform/- unordered list
        transform_unordered_list_menu_item = Gtk.MenuItem(label=parent_window.voc["unordered-list"])
        transform_menu_item.get_submenu().append(transform_unordered_list_menu_item)
        transform_unordered_list_menu_item.connect("activate", parent_window.transform_text, "unordered")

        # Transform/1. ordered list
        transform_ordered_list_menu_item = Gtk.MenuItem(label=parent_window.voc["ordered-list"])
        transform_menu_item.get_submenu().append(transform_ordered_list_menu_item)
        transform_ordered_list_menu_item.connect("activate", parent_window.transform_text, "ordered")

        # Transform/First word to the end
        first_to_end_item = Gtk.MenuItem(label=parent_window.voc["first-to-end"])
        transform_menu_item.get_submenu().append(first_to_end_item)
        first_to_end_item.connect("activate", parent_window.transform_text, "first-to-end")

        # TransformLast word to the beginning
        last_to_beginning_item = Gtk.MenuItem(label=parent_window.voc["last-to-beginning"])
        transform_menu_item.get_submenu().append(last_to_beginning_item)
        last_to_beginning_item.connect("activate", parent_window.transform_text, "last-to-beginning")

        # Merge lines
        merge_lines_menu_item = Gtk.MenuItem(label=parent_window.voc["merge-rows"])
        transform_menu_item.get_submenu().append(merge_lines_menu_item)
        merge_lines_menu_item.connect("activate", parent_window.transform_text, "merge-rows")

        separator = Gtk.SeparatorMenuItem()
        edit_menu.append(separator)

        # Preferences menu item
        preferences_menu_item = Gtk.MenuItem(label=parent_window.voc["preferences"])
        edit_menu.append(preferences_menu_item)
        preferences_menu_item.connect("activate", parent_window.show_preferences)

        # View menu
        view_menu = Gtk.Menu()
        view_menu_item = Gtk.MenuItem(label=parent_window.voc["view"])
        view_menu_item.set_submenu(view_menu)

        # Line numbers menu item
        self.line_numbers_menu_item = Gtk.CheckMenuItem(parent_window.voc["line-numbers"])
        view_menu.append(self.line_numbers_menu_item)
        self.line_numbers_menu_item.set_active(self.settings["view-line-numbers"])
        self.line_numbers_menu_item.connect("toggled", parent_window.toggle_line_numbers)

        # Whitespaces menu item
        whitespaces_menu_item = Gtk.CheckMenuItem(parent_window.voc["whitespaces"])
        view_menu.append(whitespaces_menu_item)
        whitespaces_menu_item.set_active(self.settings["whitespaces"])
        whitespaces_menu_item.connect("toggled", parent_window.toggle_whitespaces)

        # Highlight current row item
        self.highlight_current_row_menu_item = Gtk.CheckMenuItem(parent_window.voc["highlight-current-row"])
        view_menu.append(self.highlight_current_row_menu_item)
        self.highlight_current_row_menu_item.set_active(self.settings["highlight-current-row"])
        self.highlight_current_row_menu_item.connect("toggled", parent_window.toggle_highlight_current_row)

        # Highlight matching brackets
        self.highlight_matching_brackets_menu_item = Gtk.CheckMenuItem(parent_window.voc["highlight-matching-brackets"])
        view_menu.append(self.highlight_matching_brackets_menu_item)
        self.highlight_matching_brackets_menu_item.set_active(self.settings["highlight-matching-brackets"])
        self.highlight_matching_brackets_menu_item.connect("toggled", parent_window.toggle_highlight_matching_brackets)

        # Wrap menu item
        self.wrap_menu_item = Gtk.CheckMenuItem(parent_window.voc["wrap-lines"])
        view_menu.append(self.wrap_menu_item)
        self.wrap_menu_item.set_active(self.settings["wrap-lines"])
        self.wrap_menu_item.connect("toggled", parent_window.toggle_line_wrap)

        # Syntax menu item
        syntax_menu_item = Gtk.MenuItem(label=parent_window.voc["syntax-highlight"])
        view_menu.append(syntax_menu_item)
        view_menu.connect("show", add_syntax_menu, syntax_menu_item, self.parent_window)

        # Tools menu
        tools_menu = Gtk.Menu()
        tools_menu.connect("show", self.update_menu_items_sensitivity)
        tools_menu_item = Gtk.MenuItem(label=parent_window.voc["tools"])
        tools_menu_item.set_submenu(tools_menu)

        # Sanitize menu item
        self.sanitize_menu_item = Gtk.MenuItem(label=parent_window.voc["web-cleanup"])
        tools_menu.append(self.sanitize_menu_item)
        self.sanitize_menu_item.connect("show", self.update_menu_items_sensitivity)
        self.sanitize_menu_item.connect("activate", parent_window.sanitize_text)
        self.sanitize_menu_item.set_can_focus(False)

        # Sort menu item
        sort_menu_item = Gtk.MenuItem(label=parent_window.voc["sort-rows"])
        tools_menu.append(sort_menu_item)

        sort_menu = Gtk.Menu()
        sort_asc_menu_item = Gtk.MenuItem(label=parent_window.voc["ascending"])
        sort_menu.append(sort_asc_menu_item)
        sort_asc_menu_item.connect("activate", parent_window.transform_text, "sort-asc")
        sort_desc_menu_item = Gtk.MenuItem(label=parent_window.voc["descending"])
        sort_menu.append(sort_desc_menu_item)
        sort_desc_menu_item.connect("activate", parent_window.transform_text, "sort-desc")
        sort_menu_item.set_submenu(sort_menu)

        remove_empty_rows_item = Gtk.MenuItem(label=parent_window.voc["remove-empty-rows"])
        tools_menu.append(remove_empty_rows_item)
        remove_empty_rows_item.connect("activate", parent_window.transform_text, "remove-empty-rows")

        remove_non_ascii_item = Gtk.MenuItem(label=parent_window.voc["remove-non-ascii"])
        tools_menu.append(remove_non_ascii_item)
        remove_non_ascii_item.connect("activate", parent_window.transform_text, "remove-non-ascii")

        # Help menu
        help_menu = Gtk.Menu()
        help_menu_item = Gtk.MenuItem(label=parent_window.voc["help"])
        help_menu_item.set_submenu(help_menu)

        # About menu item
        about_menu_item = Gtk.MenuItem(label=parent_window.voc["about"])
        help_menu.append(about_menu_item)
        about_menu_item.connect("activate", parent_window.show_about)

        # Append menu items to the menu bar
        self.append(file_menu_item)
        self.append(edit_menu_item)
        self.append(view_menu_item)
        self.append(tools_menu_item)
        self.append(help_menu_item)

        self.update_menu_items_sensitivity()

    def update_menu_items_sensitivity(self, *args):
        self.undo_menu_item.set_sensitive(self.parent_window.buffer.can_undo())
        self.redo_menu_item.set_sensitive(self.parent_window.buffer.can_redo())
        self.new_menu_item.set_sensitive(self.parent_window.buffer.get_char_count() > 0)
        self.sanitize_menu_item.set_sensitive(self.parent_window.buffer.get_char_count() > 0)


class CustomMenuItem(Gtk.MenuItem):
    def __init__(self, label, desc=""):
        super().__init__()
        hbox = Gtk.Box(Gtk.Orientation.HORIZONTAL, spacing=0)
        self.add(hbox)
        label = Gtk.Label.new(label)
        label.set_property("xalign", 0)
        label.set_property("yalign", 0.5)
        hbox.pack_start(label, True, True, 0)
        if desc:
            desc = Gtk.Label(label=desc)
            desc.set_property("xalign", 1)
            desc.set_property("yalign", 0.5)
            hbox.pack_start(desc, True, True, 0)


def add_recent_menu(widget, parent_item, parent_window):
    menu = Gtk.Menu()
    recent_file = os.path.join(config_dir(), "recent")
    recent_paths = load_text_file(recent_file).splitlines() if os.path.isfile(recent_file) else []
    for path in recent_paths:
        item = Gtk.MenuItem(label=path)
        item.connect("activate", parent_window.load_file, path)
        menu.append(item)
        menu.show_all()
    parent_item.set_submenu(menu)


def add_syntax_menu(widget, parent_item, parent_window):
    menu = Gtk.Menu()
    item = Gtk.MenuItem(label=parent_window.voc["plain-text"])
    item.connect("activate", parent_window.set_syntax, "none")
    menu.append(item)
    for key in parent_window.syntax_dict:
        item = Gtk.MenuItem(label=parent_window.syntax_dict[key])
        item.connect("activate", parent_window.set_syntax, key)
        menu.append(item)
    menu.show_all()
    parent_item.set_submenu(menu)


class ButtonBar(Gtk.Box):
    def __init__(self, parent_window, dir_name):
        super().__init__(orientation=Gtk.Orientation.HORIZONTAL, spacing=0)
        self.settings = parent_window.settings
        self.icons_path = os.path.join(dir_name, "icons", self.settings["icon-set"])

        btn_new = self.create_button("file-new.svg")
        btn_new.set_tooltip_text(parent_window.voc["new"])
        self.pack_start(btn_new, False, False, 0)
        btn_new.connect("clicked", parent_window.new_file)

        btn_open = self.create_button("file-open.svg")
        btn_open.set_tooltip_text(parent_window.voc["open"])
        self.pack_start(btn_open, False, False, 0)
        btn_open.connect("clicked", parent_window.open_file)

        btn_save = self.create_button("file-save.svg")
        btn_save.set_tooltip_text(parent_window.voc["save"])
        self.pack_start(btn_save, False, False, 0)
        btn_save.connect("clicked", parent_window.save_file)

        btn_save_as = self.create_button("file-save-as.svg")
        btn_save_as.set_tooltip_text(parent_window.voc["save-as"])
        self.pack_start(btn_save_as, False, False, 0)
        btn_save_as.connect("clicked", parent_window.save_file_as)

        self.pack_start(self.create_separator(), False, False, 0)

        btn_undo = self.create_button("edit-undo.svg")
        btn_undo.set_tooltip_text(parent_window.voc["undo"])
        self.pack_start(btn_undo, False, False, 0)
        btn_undo.connect("clicked", parent_window.undo)

        btn_redo = self.create_button("edit-redo.svg")
        btn_redo.set_tooltip_text(parent_window.voc["redo"])
        self.pack_start(btn_redo, False, False, 0)
        btn_redo.connect("clicked", parent_window.redo)

        btn_copy = self.create_button("edit-copy.svg")
        btn_copy.set_tooltip_text(parent_window.voc["copy"])
        self.pack_start(btn_copy, False, False, 0)
        btn_copy.connect("clicked", parent_window.copy_text)

        btn_cut = self.create_button("edit-cut.svg")
        btn_cut.set_tooltip_text(parent_window.voc["cut"])
        self.pack_start(btn_cut, False, False, 0)
        btn_cut.connect("clicked", parent_window.cut_text)

        btn_paste = self.create_button("edit-paste.svg")
        btn_paste.set_tooltip_text(parent_window.voc["paste"])
        self.pack_start(btn_paste, False, False, 0)
        btn_paste.connect("clicked", parent_window.paste_text)

        self.pack_start(self.create_separator(), False, False, 0)

        btn_sentence = self.create_button("as-in-sentence.svg")
        btn_sentence.set_tooltip_text(parent_window.voc["as-in-sentence"])
        self.pack_start(btn_sentence, False, False, 0)
        btn_sentence.connect("clicked", parent_window.transform_text, "sentence")

        btn_title = self.create_button("as-in-title.svg")
        btn_title.set_tooltip_text(parent_window.voc["as-in-title"])
        self.pack_start(btn_title, False, False, 0)
        btn_title.connect("clicked", parent_window.transform_text, "title")

        btn_uppercase = self.create_button("uppercase.svg")
        btn_uppercase.set_tooltip_text(parent_window.voc["uppercase"])
        self.pack_start(btn_uppercase, False, False, 0)
        btn_uppercase.connect("clicked", parent_window.transform_text, "uppercase")

        btn_lowercase = self.create_button("lowercase.svg")
        btn_lowercase.set_tooltip_text(parent_window.voc["lowercase"])
        self.pack_start(btn_lowercase, False, False, 0)
        btn_lowercase.connect("clicked", parent_window.transform_text, "lowercase")

        btn_camelcase = self.create_button("camel-case.svg")
        btn_camelcase.set_tooltip_text(parent_window.voc["camel-case"])
        self.pack_start(btn_camelcase, False, False, 0)
        btn_camelcase.connect("clicked", parent_window.transform_text, "camelcase")

        btn_snakecase = self.create_button("snake-case.svg")
        btn_snakecase.set_tooltip_text(parent_window.voc["snake-case"])
        self.pack_start(btn_snakecase, False, False, 0)
        btn_snakecase.connect("clicked", parent_window.transform_text, "snakecase")

        btn_kebabcase = self.create_button("kebab-case.svg")
        btn_kebabcase.set_tooltip_text(parent_window.voc["kebab-case"])
        self.pack_start(btn_kebabcase, False, False, 0)
        btn_kebabcase.connect("clicked", parent_window.transform_text, "kebabcase")

        btn_unordered = self.create_button("unordered-list.svg")
        btn_unordered.set_tooltip_text(parent_window.voc["unordered-list"])
        self.pack_start(btn_unordered, False, False, 0)
        btn_unordered.connect("clicked", parent_window.transform_text, "unordered")

        btn_ordered = self.create_button("ordered-list.svg")
        btn_ordered.set_tooltip_text(parent_window.voc["ordered-list"])
        self.pack_start(btn_ordered, False, False, 0)
        btn_ordered.connect("clicked", parent_window.transform_text, "ordered")

        btn_first = self.create_button("first-to-end.svg")
        btn_first.set_tooltip_text(parent_window.voc["first-to-end"])
        self.pack_start(btn_first, False, False, 0)
        btn_first.connect("clicked", parent_window.transform_text, "first-to-end")

        btn_last = self.create_button("last-to-beginning.svg")
        btn_last.set_tooltip_text(parent_window.voc["last-to-beginning"])
        self.pack_start(btn_last, False, False, 0)
        btn_last.connect("clicked", parent_window.transform_text, "last-to-beginning")

        btn_merge = self.create_button("merge-rows.svg")
        btn_merge.set_tooltip_text(parent_window.voc["merge-rows"])
        self.pack_start(btn_merge, False, False, 0)
        btn_merge.connect("clicked", parent_window.transform_text, "merge-rows")

        self.pack_start(self.create_separator(), False, False, 0)

        btn_web = self.create_button("web-cleanup.svg")
        btn_web.set_tooltip_text(parent_window.voc["web-cleanup"])
        self.pack_start(btn_web, False, False, 0)
        btn_web.connect("clicked", parent_window.sanitize_text)

        btn_sort_ascending = self.create_button("sort-ascending.svg")
        btn_sort_ascending.set_tooltip_text(parent_window.voc["ascending"])
        self.pack_start(btn_sort_ascending, False, False, 0)
        btn_sort_ascending.connect("clicked", parent_window.transform_text, "sort-asc")

        btn_sort_descending = self.create_button("sort-descending.svg")
        btn_sort_descending.set_tooltip_text(parent_window.voc["descending"])
        self.pack_start(btn_sort_descending, False, False, 0)
        btn_sort_descending.connect("clicked", parent_window.transform_text, "sort-desc")

        btn_remove_empty = self.create_button("remove-empty-rows.svg")
        btn_remove_empty.set_tooltip_text(parent_window.voc["remove-empty-rows"])
        self.pack_start(btn_remove_empty, False, False, 0)
        btn_remove_empty.connect("clicked", parent_window.transform_text, "remove-empty-rows")

        btn_remove_non_ascii = self.create_button("remove-non-ascii.svg")
        btn_remove_non_ascii.set_tooltip_text(parent_window.voc["remove-non-ascii"])
        self.pack_start(btn_remove_non_ascii, False, False, 0)
        btn_remove_non_ascii.connect("clicked", parent_window.transform_text, "remove-non-ascii")

        self.show_all()

    def create_button(self, icon_name):
        btn = Gtk.Button()
        btn.set_image(self.create_image(icon_name))
        btn.set_property("name", "bar-button")
        return btn

    def create_separator(self):
        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(os.path.join(self.icons_path, "separator.svg"),
                                                        self.settings["icon-size"] / 2, self.settings["icon-size"])
        img = Gtk.Image.new_from_pixbuf(pixbuf)
        return img

    def create_image(self, icon_name):
        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(os.path.join(self.icons_path, icon_name),
                                                        self.settings["icon-size"], self.settings["icon-size"])
        return Gtk.Image.new_from_pixbuf(pixbuf)


class SanitizationDialog(Gtk.Window):
    def __init__(self, parent_window, buffer):
        super().__init__(title="Sanitize Text")
        self.set_transient_for(parent_window)
        self.set_modal(True)
        self.parent_window = parent_window
        self.settings = parent_window.settings

        self.connect("key-release-event", self.handle_keyboard_release)

        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        vbox.set_property("margin", 12)
        self.add(vbox)

        self.sanitize_hyphens = Gtk.CheckButton(label=parent_window.voc["hyphens"])
        self.sanitize_hyphens.set_active(self.settings["sanitize-hyphens"])
        self.sanitize_hyphens.connect("toggled", self.switch_settings_key, "sanitize-hyphens")
        vbox.pack_start(self.sanitize_hyphens, False, False, 0)

        self.sanitize_quotes = Gtk.CheckButton(label=parent_window.voc["quotes"])
        self.sanitize_quotes.set_active(self.settings["sanitize-quotes"])
        self.sanitize_quotes.connect("toggled", self.switch_settings_key, "sanitize-quotes")
        vbox.pack_start(self.sanitize_quotes, False, False, 0)

        self.sanitize_punctuation_marks = Gtk.CheckButton(label=parent_window.voc["punctuation-marks"])
        self.sanitize_punctuation_marks.set_active(self.settings["sanitize-punctuation-marks"])
        self.sanitize_punctuation_marks.connect("toggled", self.switch_settings_key, "sanitize-punctuation-marks")
        vbox.pack_start(self.sanitize_punctuation_marks, False, False, 0)

        self.sanitize_spaces = Gtk.CheckButton(label=parent_window.voc["spaces"])
        self.sanitize_spaces.set_active(self.settings["sanitize-spaces"])
        self.sanitize_spaces.connect("toggled", self.switch_settings_key, "sanitize-spaces")
        vbox.pack_start(self.sanitize_spaces, False, False, 0)

        self.add_spaces_after = Gtk.CheckButton(label=parent_window.voc["add-spaces-after-punctuation"])
        self.add_spaces_after.set_active(self.settings["sanitize-add-spaces-after-punctuation"])
        self.add_spaces_after.connect("toggled", self.switch_settings_key,
                                      "sanitize-add-spaces-after-punctuation")
        vbox.pack_start(self.add_spaces_after, False, False, 0)

        self.sanitize_eol = Gtk.CheckButton(label=parent_window.voc["eol-chars"])
        self.sanitize_eol.set_active(self.settings["sanitize-eol"])
        self.sanitize_eol.connect("toggled", self.switch_settings_key, "sanitize-eol")
        vbox.pack_start(self.sanitize_eol, False, False, 0)

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        hbox.set_property("margin-top", 12)
        vbox.pack_start(hbox, False, False, 0)

        button = Gtk.Button(label=parent_window.voc["sanitize"])
        hbox.pack_end(button, False, False, 0)
        button.connect("clicked", self.sanitize_text, buffer)

        button = Gtk.Button(label=parent_window.voc["close"])
        hbox.pack_end(button, False, False, 0)
        button.connect("clicked", lambda x: self.destroy())

        self.show_all()

    def handle_keyboard_release(self, widget, event):
        if event.keyval == Gdk.KEY_Escape:
            self.destroy()

    def switch_settings_key(self, chekbox, key):
        if key in self.settings:
            self.settings[key] = chekbox.get_active()
            save_settings(self.settings)
        else:
            print(f"Key '{key}' not found in settings")

    def sanitize_text(self, widget, buffer):
        text, start_idx, end_idx = selected_text(buffer)

        if self.settings["sanitize-hyphens"]:
            text = sanitize_hyphens(text, start_idx, end_idx)

        if self.settings["sanitize-quotes"]:
            text = sanitize_quotes(text, start_idx, end_idx)

        if self.settings["sanitize-punctuation-marks"]:
            text = sanitize_punctuation_marks(text, start_idx, end_idx)

        if self.settings["sanitize-add-spaces-after-punctuation"]:
            text = add_spaces_after_punctuation_marks(text, start_idx, end_idx)

        if self.settings["sanitize-spaces"]:
            text = sanitize_spaces(text, start_idx, end_idx, self.settings["tab-mode"] == "spaces",
                                   self.settings["tab-width"])

        if self.settings["sanitize-eol"]:
            text = sanitize_eol(text, start_idx, end_idx)

        self.parent_window.update_text(text)
        self.destroy()


class AboutWindow(Gtk.AboutDialog):
    def __init__(self, parent_window):
        super().__init__()
        self.set_transient_for(parent_window)
        self.set_program_name("Typobuster")
        self.set_copyright(parent_window.voc["copyright"])
        self.set_comments(parent_window.voc["description"])
        self.set_website("https://github.com/nwg-piotr/typobuster")
        self.set_website_label("GitHub repository")
        self.set_authors(["Piotr Miller"])
        self.set_translator_credits("Marcelo dos Santos Mafra (pt_BR)\nPiotr Miller (pl_PL)")
        self.set_license_type(Gtk.License.GPL_3_0)
        self.set_version(f'{parent_window.voc["version"]}: {__version__}')

        # Load icon by name from the system theme
        icon_theme = Gtk.IconTheme.get_default()
        icon_name = "typobuster"
        pixbuf = icon_theme.load_icon(icon_name, 128, 0)  # 128x128 size
        self.set_logo(pixbuf)

        self.set_keep_above(True)
        self.connect("response", lambda x, y: self.destroy())
        self.show_all()


class PreferencesDialog(Gtk.Dialog):
    def __init__(self, parent):
        super().__init__(title="Preferences", transient_for=parent, modal=True)

        self.grid = Gtk.Grid(margin_top=10, margin_bottom=10, margin_start=10, margin_end=10, column_spacing=10,
                             row_spacing=10)
        self.get_content_area().pack_start(self.grid, False, False, 0)

        # Theme Selector
        self.theme_label = Gtk.Label(label=parent.voc["gtk-theme"], halign=Gtk.Align.START)
        self.grid.attach(self.theme_label, 0, 0, 1, 1)

        self.theme_combo = Gtk.ComboBoxText()
        self.theme_combo.append("", parent.voc["system-default"])
        for name in get_theme_names():
            self.theme_combo.append(name, name)

        if parent.settings["gtk-theme-name"]:
            self.theme_combo.set_active_id(parent.settings["gtk-theme-name"])
        else:
            self.theme_combo.set_active_id("")

        self.theme_combo.connect("changed", parent.on_theme_changed)
        self.grid.attach(self.theme_combo, 1, 0, 1, 1)

        # Font Selector
        self.font_size_label = Gtk.Label(label=parent.voc["editor-font"], halign=Gtk.Align.START)
        self.grid.attach(self.font_size_label, 0, 1, 1, 1)

        self.font_chooser_btn = Gtk.FontButton()
        self.font_chooser_btn.set_use_font(True)
        self.font_chooser_btn.set_font_name(parent.settings["gtk-font-name"])
        self.font_chooser_btn.connect("font-set", parent.on_font_selected)
        self.grid.attach(self.font_chooser_btn, 1, 1, 1, 1)

        self.right_margin_label = Gtk.Label(label=parent.voc["right-margin-position"], halign=Gtk.Align.START)
        self.grid.attach(self.right_margin_label, 0, 2, 1, 1)

        right_margin_sb = Gtk.SpinButton.new_with_range(1, 1024.0, 1)
        right_margin_sb.set_value(parent.settings["right-margin-position"])
        self.grid.attach(right_margin_sb, 1, 2, 1, 1)
        right_margin_sb.connect("value-changed", parent.on_margin_position_selected)

        right_margin_cb = Gtk.CheckButton.new_with_label(parent.voc["show"])
        right_margin_cb.set_active(parent.settings["right-margin-show"])
        self.grid.attach(right_margin_cb, 2, 2, 1, 1)
        right_margin_cb.connect("toggled", parent.on_right_margin_toggled)

        self.icon_size_label = Gtk.Label(label=parent.voc["button-bar-icon-size"], halign=Gtk.Align.START)
        self.grid.attach(self.icon_size_label, 0, 3, 1, 1)

        icon_bar_cb = Gtk.CheckButton.new_with_label(parent.voc["show"])
        icon_bar_cb.set_active(parent.settings["show-bar"])
        self.grid.attach(icon_bar_cb, 2, 3, 1, 1)
        icon_bar_cb.connect("toggled", parent.on_bar_show_toggled)

        icon_size_sb = Gtk.SpinButton.new_with_range(16, 128, 4)
        icon_size_sb.set_value(parent.settings["icon-size"])
        self.grid.attach(icon_size_sb, 1, 3, 1, 1)
        icon_size_sb.connect("value-changed", parent.on_icon_size_selected)

        icon_set_label = Gtk.Label(label=parent.voc["button-bar-icon-set"], halign=Gtk.Align.START)
        self.grid.attach(icon_set_label, 0, 4, 1, 1)

        icon_set_combo = Gtk.ComboBoxText()
        icon_set_combo.append("light", parent.voc["light"])
        icon_set_combo.append("dark", parent.voc["dark"])
        icon_set_combo.set_active_id(parent.settings["icon-set"])
        icon_set_combo.connect("changed", parent.on_icon_set_changed)
        self.grid.attach(icon_set_combo, 1, 4, 1, 1)

        self.tab_width_label = Gtk.Label(label=parent.voc["tab-width"], halign=Gtk.Align.START)
        self.grid.attach(self.tab_width_label, 0, 5, 1, 1)

        tab_width_sb = Gtk.SpinButton.new_with_range(1, 32.0, 1)
        tab_width_sb.set_value(parent.settings["tab-width"])
        tab_width_sb.connect("value-changed", parent.on_tab_with_selected)
        self.grid.attach(tab_width_sb, 1, 5, 1, 1)

        self.tab_mode_label = Gtk.Label(label=parent.voc["tab-mode"], halign=Gtk.Align.START)
        self.grid.attach(self.tab_mode_label, 0, 6, 1, 1)

        self.tab_mode_combo = Gtk.ComboBoxText()
        self.tab_mode_combo.append("tabs", parent.voc["insert-tabs"])
        self.tab_mode_combo.append("spaces", parent.voc["insert-spaces"])
        self.tab_mode_combo.set_active_id(parent.settings["tab-mode"])
        self.tab_mode_combo.connect("changed", parent.on_tab_mode_changed)
        self.grid.attach(self.tab_mode_combo, 1, 6, 1, 1)

        self.auto_indent_cb = Gtk.CheckButton(label=parent.voc["auto-indent"])
        self.auto_indent_cb.set_active(parent.settings["auto-indent"])
        self.auto_indent_cb.connect("toggled", parent.on_auto_indent_changed)
        self.grid.attach(self.auto_indent_cb, 0, 7, 1, 1)

        self.spell_check_cb = Gtk.CheckButton(label=parent.voc["spell-check"])
        self.spell_check_cb.set_sensitive(parent.gspell_available)
        if not parent.gspell_available:
            self.spell_check_cb.set_tooltip_text(parent.voc["gspell-missing"])
        self.spell_check_cb.set_active(parent.settings["gspell-enable"])
        self.spell_check_cb.connect("toggled", parent.on_spell_check_switched)
        self.grid.attach(self.spell_check_cb, 1, 7, 1, 1)

        self.stats_cb = Gtk.CheckButton(label=parent.voc["show-stats"])
        self.stats_cb.set_active(parent.settings["show-stats"])
        self.stats_cb.connect("toggled", parent.on_stats_cb_toggled)
        self.grid.attach(self.stats_cb, 0, 8, 1, 1)

        self.change_cb = Gtk.CheckButton(label=parent.voc["show-change-mark"])
        self.change_cb.set_active(parent.settings["show-change"])
        self.change_cb.connect("toggled", parent.on_change_cb_toggled)
        self.grid.attach(self.change_cb, 1, 8, 1, 1)

        # OK Button
        hbox = Gtk.Box(Gtk.Orientation.HORIZONTAL, 0)
        self.grid.attach(hbox, 0, 9, 3, 1)
        self.ok_button = Gtk.Button(label=parent.voc["close"])
        self.ok_button.connect("clicked", lambda x: self.close())
        hbox.pack_end(self.ok_button, False, False, 0)

        self.show_all()


class SearchBar(Gtk.Box):
    def __init__(self, parent_window):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.HORIZONTAL, spacing=3)
        self.set_property("margin", 3)
        self.parent_window = parent_window
        self.matches = None
        self.match_idx = -1

        self.search_entry = Gtk.SearchEntry()
        self.pack_start(self.search_entry, False, False, 1)
        self.search_entry.set_property("name", "searchentry")
        self.search_entry.set_tooltip_text(parent_window.voc["search"])
        self.search_entry.connect("search-changed", self.on_search_changed, parent_window.buffer)
        self.search_entry.connect("key-release-event", self.handle_keyboard_release)

        btn = Gtk.Button.new_from_icon_name("go-up-symbolic", Gtk.IconSize.MENU)
        self.pack_start(btn, False, False, 0)
        btn.connect("clicked", self.highlight_match, "up")

        btn = Gtk.Button.new_from_icon_name("go-down-symbolic", Gtk.IconSize.MENU)
        self.pack_start(btn, False, False, 0)
        btn.connect("clicked", self.highlight_match, "down")

        self.replace_entry = Gtk.Entry()
        self.replace_entry.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY, "edit-find-replace-symbolic")
        self.replace_entry.set_tooltip_text(parent_window.voc["replace-with"])
        self.replace_entry.connect("icon-press", lambda x, y, z: self.replace_entry.set_text(""))
        self.pack_start(self.replace_entry, False, False, 0)

        btn = Gtk.Button.new_from_icon_name("emblem-ok-symbolic", Gtk.IconSize.MENU)
        self.pack_start(btn, False, False, 0)
        btn.connect("clicked", self.replace)

        self.pos_lbl = Gtk.Label.new(f'{parent_window.voc["row"]}: 1 {parent_window.voc["column"]}: 0')
        self.pack_end(self.pos_lbl, False, False, 3)

        self.stat_lbl = Gtk.Label.new("0")
        self.pack_end(self.stat_lbl, False, False, 3)

        if parent_window.settings["syntax"] == "none":
            s_lbl = parent_window.voc["plain-text"]
        else:
            s_lbl = parent_window.syntax_dict[parent_window.settings["syntax"]]
        self.syntax_lbl = Gtk.Label.new(s_lbl)
        self.pack_end(self.syntax_lbl, False, False, 0)

        self.change_lbl = Gtk.Label.new("")
        self.pack_end(self.change_lbl, False, False, 0)

        self.show_all()

    def handle_keyboard_release(self, widget, event):
        if event.keyval == Gdk.KEY_Return:
            self.highlight_match(widget, "down")

    def on_search_changed(self, widget, buffer):
        self.match_idx = -1
        phrase = self.search_entry.get_text()
        if phrase:
            start = buffer.get_start_iter()
            end = buffer.get_end_iter()
            text = buffer.get_text(start, end, True)

            if phrase in text:
                self.search_entry.set_property("name", "searchentry")
            else:
                self.search_entry.set_property("name", "searchentry-error")

            self.matches = list(re.finditer(re.escape(phrase), text))

    def highlight_match(self, widget, direction):
        if self.search_entry.get_text():
            if direction == "down":
                if self.match_idx == -1:
                    self.match_idx = 0

                start, end = self.matches[self.match_idx].span()
                self.parent_window.select_range(start, end)
                iter_start = self.parent_window.buffer.get_iter_at_offset(start)
                self.parent_window.source_view.scroll_to_iter(iter_start, 0.2, False, 0.5, 0.5)

                if self.match_idx < len(self.matches) - 1:
                    self.match_idx += 1
                else:
                    self.match_idx = 0

            elif direction == "up":
                if self.match_idx == -1:
                    self.match_idx = len(self.matches) - 1

                start, end = self.matches[self.match_idx].span()
                self.parent_window.select_range(start, end)
                iter_start = self.parent_window.buffer.get_iter_at_offset(start)
                self.parent_window.source_view.scroll_to_iter(iter_start, 0.2, False, 0.5, 0.5)

                if self.match_idx > 0:
                    self.match_idx -= 1
                else:
                    self.match_idx = len(self.matches) - 1

    def replace(self, widget):
        old = self.search_entry.get_text()
        new = self.replace_entry.get_text()
        if old:
            self.parent_window.replace(old, new)

    def clear(self):
        self.search_entry.set_text("")
        self.replace_entry.set_text("")
